#!/usr/bin/env python
import os
import json
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

from scipy.stats import norm
from itertools import product
import anndata
import numpy as np
import pandas as pd
import scanpy as sc

from scipy.sparse import issparse
from sklearn.metrics import calinski_harabasz_score

"""
HashSolo script provides a probabilistic cell hashing demultiplexing method
which generates a noise distribution and signal distribution for
each hashing barcode from empirically observed counts. These distributions
are updates from the global signal and noise barcode distributions, which
helps in the setting where not many cells are observed. Signal distributions
for a hashing barcode are estimated from samples where that hashing barcode
has the highest count. Noise distributions for a hashing barcode are estimated
from samples where that hashing barcode is one the k-2 lowest barcodes, where
k is the number of barcodes. A doublet should then have its two highest
barcode counts most likely coming from a signal distribution for those barcodes.
A singlet should have its highest barcode from a signal distribution, and its
second highest barcode from a noise distribution. A negative two highest
barcodes should come from noise distributions. We test each of these
hypotheses in a bayesian fashion, and select the most probable hypothesis.
"""


def _calculate_log_likelihoods(data, number_of_noise_barcodes):
    """Calculate log likelihoods for each hypothesis, negative, singlet, doublet

    Parameters
    ----------
    data : np.ndarray
        cells by hashing counts matrix
    number_of_noise_barcodes : int,
        number of barcodes to used to calculated noise distribution
    Returns
    -------
    log_likelihoods_for_each_hypothesis : np.ndarray
        a 2d np.array log likelihood of each hypothesis
    all_indices
    counter_to_barcode_combo
    """

    def gaussian_updates(data, mu_o, std_o):
        """Update parameters of your gaussian
        https://www.cs.ubc.ca/~murphyk/Papers/bayesGauss.pdf
        Parameters
        ----------
        data : np.array
            1-d array of counts
        mu_o : float,
            global mean for hashing count distribution
        std_o : float,
            global std for hashing count distribution
        Returns
        -------
        float
            mean of gaussian
        float
            std of gaussian
        """
        lam_o = 1 / (std_o ** 2)
        n = len(data)
        lam = 1 / np.var(data) if len(data) > 1 else lam_o
        lam_n = lam_o + n * lam
        mu_n = (
            (np.mean(data) * n * lam + mu_o * lam_o) / lam_n if len(data) > 0 else mu_o
        )
        return mu_n, (1 / (lam_n / (n + 1))) ** (1 / 2)

    eps = 1e-15
    # probabilites for negative, singlet, doublets
    log_likelihoods_for_each_hypothesis = np.zeros((data.shape[0], 3))

    all_indices = np.empty(data.shape[0])
    num_of_barcodes = data.shape[1]
    number_of_non_noise_barcodes = (
        num_of_barcodes - number_of_noise_barcodes
        if number_of_noise_barcodes is not None
        else 2
    )
    num_of_noise_barcodes = num_of_barcodes - number_of_non_noise_barcodes

    # assume log normal
    data = np.log(data + 1)
    data_arg = np.argsort(data, axis=1)
    data_sort = np.sort(data, axis=1)

    # global signal and noise counts useful for when we have few cells
    # barcodes with the highest number of counts are assumed to be a true signal
    # barcodes with rank < k are considered to be noise
    global_signal_counts = np.ravel(data_sort[:, -1])
    global_noise_counts = np.ravel(data_sort[:, :-number_of_non_noise_barcodes])
    global_mu_signal_o, global_sigma_signal_o = np.mean(global_signal_counts), np.std(
        global_signal_counts
    )
    global_mu_noise_o, global_sigma_noise_o = np.mean(global_noise_counts), np.std(
        global_noise_counts
    )

    noise_params_dict = {}
    signal_params_dict = {}

    # for each barcode get  empirical noise and signal distribution parameterization
    for x in np.arange(num_of_barcodes):
        sample_barcodes = data[:, x]
        sample_barcodes_noise_idx = np.where(data_arg[:, :num_of_noise_barcodes] == x)[
            0
        ]
        sample_barcodes_signal_idx = np.where(data_arg[:, -1] == x)

        # get noise and signal counts
        noise_counts = sample_barcodes[sample_barcodes_noise_idx]
        signal_counts = sample_barcodes[sample_barcodes_signal_idx]

        # get parameters of distribution, assuming lognormal do update from global values
        noise_param = gaussian_updates(
            noise_counts, global_mu_noise_o, global_sigma_noise_o
        )
        signal_param = gaussian_updates(
            signal_counts, global_mu_signal_o, global_sigma_signal_o
        )
        noise_params_dict[x] = noise_param
        signal_params_dict[x] = signal_param

    counter_to_barcode_combo = {}
    counter = 0

    # for each combination of noise and signal barcode calculate probiltiy of in silico and real cell hypotheses
    for noise_sample_idx, signal_sample_idx in product(
        np.arange(num_of_barcodes), np.arange(num_of_barcodes)
    ):
        signal_subset = data_arg[:, -1] == signal_sample_idx
        noise_subset = data_arg[:, -2] == noise_sample_idx
        subset = signal_subset & noise_subset
        if sum(subset) == 0:
            continue

        indices = np.where(subset)[0]
        barcode_combo = "_".join([str(noise_sample_idx), str(signal_sample_idx)])
        all_indices[np.where(subset)[0]] = counter
        counter_to_barcode_combo[counter] = barcode_combo
        counter += 1
        noise_params = noise_params_dict[noise_sample_idx]
        signal_params = signal_params_dict[signal_sample_idx]

        # calculate probabilties for each hypothesis for each cell
        data_subset = data[subset]
        log_signal_signal_probs = np.log(
            norm.pdf(
                data_subset[:, signal_sample_idx],
                *signal_params[:-2],
                loc=signal_params[-2],
                scale=signal_params[-1]
            )
            + eps
        )
        signal_noise_params = signal_params_dict[noise_sample_idx]
        log_noise_signal_probs = np.log(
            norm.pdf(
                data_subset[:, noise_sample_idx],
                *signal_noise_params[:-2],
                loc=signal_noise_params[-2],
                scale=signal_noise_params[-1]
            )
            + eps
        )

        log_noise_noise_probs = np.log(
            norm.pdf(
                data_subset[:, noise_sample_idx],
                *noise_params[:-2],
                loc=noise_params[-2],
                scale=noise_params[-1]
            )
            + eps
        )
        log_signal_noise_probs = np.log(
            norm.pdf(
                data_subset[:, signal_sample_idx],
                *noise_params[:-2],
                loc=noise_params[-2],
                scale=noise_params[-1]
            )
            + eps
        )

        probs_of_negative = np.sum(
            [log_noise_noise_probs, log_signal_noise_probs], axis=0
        )
        probs_of_singlet = np.sum(
            [log_noise_noise_probs, log_signal_signal_probs], axis=0
        )
        probs_of_doublet = np.sum(
            [log_noise_signal_probs, log_signal_signal_probs], axis=0
        )
        log_probs_list = [probs_of_negative, probs_of_singlet, probs_of_doublet]

        # each cell and each hypothesis probability
        for prob_idx, log_prob in enumerate(log_probs_list):
            log_likelihoods_for_each_hypothesis[indices, prob_idx] = log_prob
    return log_likelihoods_for_each_hypothesis, all_indices, counter_to_barcode_combo


def _calculate_bayes_rule(data, priors, number_of_noise_barcodes):
    """
    Calculate bayes rule from log likelihoods

    Parameters
    ----------
    data : np.array
        Anndata object filled only with hashing counts
    priors : list,
        a list of your prior for each hypothesis
        first element is your prior for the negative hypothesis
        second element is your prior for the singlet hypothesis
        third element is your prior for the doublet hypothesis
        We use [0.01, 0.8, 0.19] by default because we assume the barcodes
        in your cell hashing matrix are those cells which have passed QC
        in the transcriptome space, e.g. UMI counts, pct mito reads, etc.
    number_of_noise_barcodes : int
        number of barcodes to used to calculated noise distribution
    Returns
    -------
    bayes_dict_results : dict
        'most_likely_hypothesis' key is a 1d np.array of the most likely hypothesis
        'probs_hypotheses' key is a 2d np.array probability of each hypothesis
        'log_likelihoods_for_each_hypothesis' key is a 2d np.array log likelihood of each hypothesis
    """
    priors = np.array(priors)
    log_likelihoods_for_each_hypothesis, _, _ = _calculate_log_likelihoods(
        data, number_of_noise_barcodes
    )
    probs_hypotheses = (
        np.exp(log_likelihoods_for_each_hypothesis)
        * priors
        / np.sum(
            np.multiply(np.exp(log_likelihoods_for_each_hypothesis), priors), axis=1
        )[:, None]
    )
    most_likely_hypothesis = np.argmax(probs_hypotheses, axis=1)
    return {
        "most_likely_hypothesis": most_likely_hypothesis,
        "probs_hypotheses": probs_hypotheses,
        "log_likelihoods_for_each_hypothesis": log_likelihoods_for_each_hypothesis,
    }


def _get_clusters(clustering_data: anndata.AnnData, resolutions: list):
    """
    Principled cell clustering
    Parameters
    ----------
    cell_hashing_adata : anndata.AnnData
        Anndata object filled only with hashing counts
    resolutions : list
        clustering resolutions for leiden
    Returns
    -------
    np.ndarray
        leiden clustering results for each cell
    """
    sc.pp.normalize_per_cell(clustering_data, counts_per_cell_after=1e4)
    sc.pp.log1p(clustering_data)
    sc.pp.highly_variable_genes(
        clustering_data, min_mean=0.0125, max_mean=3, min_disp=0.5
    )
    clustering_data = clustering_data[:, clustering_data.var["highly_variable"]]
    sc.pp.scale(clustering_data, max_value=10)
    sc.tl.pca(clustering_data, svd_solver="arpack")
    sc.pp.neighbors(clustering_data, n_neighbors=10, n_pcs=40)
    sc.tl.umap(clustering_data)
    best_ch_score = -np.inf

    for resolution in resolutions:
        sc.tl.leiden(clustering_data, resolution=resolution)

        ch_score = calinski_harabasz_score(
            clustering_data.X, clustering_data.obs["leiden"]
        )

        if ch_score > best_ch_score:
            clustering_data.obs["best_leiden"] = clustering_data.obs["leiden"].values
            best_ch_score = ch_score
    return clustering_data.obs["best_leiden"].values


def hashsolo(
    cell_hashing_adata: anndata.AnnData,
    priors: list = [0.01, 0.8, 0.19],
    pre_existing_clusters: str = None,
    clustering_data: anndata.AnnData = None,
    resolutions: list = [0.1, 0.25, 0.5, 0.75, 1],
    number_of_noise_barcodes: int = None,
    inplace: bool = True,
):
    """Demultiplex cell hashing dataset using HashSolo method

    Parameters
    ----------
    cell_hashing_adata : anndata.AnnData
        Anndata object filled only with hashing counts
    priors : list,
        a list of your prior for each hypothesis
        first element is your prior for the negative hypothesis
        second element is your prior for the singlet hypothesis
        third element is your prior for the doublet hypothesis
        We use [0.01, 0.8, 0.19] by default because we assume the barcodes
        in your cell hashing matrix are those cells which have passed QC
        in the transcriptome space, e.g. UMI counts, pct mito reads, etc.
    clustering_data : anndata.AnnData
        transcriptional data for clustering
    resolutions : list
        clustering resolutions for leiden
    pre_existing_clusters : str
        column in cell_hashing_adata.obs for how to break up demultiplexing
    inplace : bool
        To do operation in place

    Returns
    -------
    cell_hashing_adata : AnnData
        if inplace is False returns AnnData with demultiplexing results
        in .obs attribute otherwise does is in place
    """
    if issparse(cell_hashing_adata.X):
        cell_hashing_adata.X = np.array(cell_hashing_adata.X.todense())

    if clustering_data is not None:
        print(
            "This may take awhile we are running clustering at {} different resolutions".format(
                len(resolutions)
            )
        )
        if not all(clustering_data.obs_names == cell_hashing_adata.obs_names):
            raise ValueError(
                "clustering_data and cell hashing cell_hashing_adata must have same index"
            )
        cell_hashing_adata.obs["best_leiden"] = _get_clusters(
            clustering_data, resolutions
        )

    data = cell_hashing_adata.X
    num_of_cells = cell_hashing_adata.shape[0]
    results = pd.DataFrame(
        np.zeros((num_of_cells, 6)),
        columns=[
            "most_likely_hypothesis",
            "probs_hypotheses",
            "cluster_feature",
            "negative_hypothesis_probability",
            "singlet_hypothesis_probability",
            "doublet_hypothesis_probability",
        ],
        index=cell_hashing_adata.obs_names,
    )
    if clustering_data is not None or pre_existing_clusters is not None:
        cluster_features = (
            "best_leiden" if pre_existing_clusters is None else pre_existing_clusters
        )
        unique_cluster_features = np.unique(cell_hashing_adata.obs[cluster_features])
        for cluster_feature in unique_cluster_features:
            cluster_feature_bool_vector = (
                cell_hashing_adata.obs[cluster_features] == cluster_feature
            )
            posterior_dict = _calculate_bayes_rule(
                data[cluster_feature_bool_vector], priors, number_of_noise_barcodes
            )
            results.loc[
                cluster_feature_bool_vector, "most_likely_hypothesis"
            ] = posterior_dict["most_likely_hypothesis"]
            results.loc[
                cluster_feature_bool_vector, "cluster_feature"
            ] = cluster_feature
            results.loc[
                cluster_feature_bool_vector, "negative_hypothesis_probability"
            ] = posterior_dict["probs_hypotheses"][:, 0]
            results.loc[
                cluster_feature_bool_vector, "singlet_hypothesis_probability"
            ] = posterior_dict["probs_hypotheses"][:, 1]
            results.loc[
                cluster_feature_bool_vector, "doublet_hypothesis_probability"
            ] = posterior_dict["probs_hypotheses"][:, 2]
    else:
        posterior_dict = _calculate_bayes_rule(data, priors, number_of_noise_barcodes)
        results.loc[:, "most_likely_hypothesis"] = posterior_dict[
            "most_likely_hypothesis"
        ]
        results.loc[:, "cluster_feature"] = 0
        results.loc[:, "negative_hypothesis_probability"] = posterior_dict[
            "probs_hypotheses"
        ][:, 0]
        results.loc[:, "singlet_hypothesis_probability"] = posterior_dict[
            "probs_hypotheses"
        ][:, 1]
        results.loc[:, "doublet_hypothesis_probability"] = posterior_dict[
            "probs_hypotheses"
        ][:, 2]

    cell_hashing_adata.obs["most_likely_hypothesis"] = results.loc[
        cell_hashing_adata.obs_names, "most_likely_hypothesis"
    ]
    cell_hashing_adata.obs["cluster_feature"] = results.loc[
        cell_hashing_adata.obs_names, "cluster_feature"
    ]
    cell_hashing_adata.obs["negative_hypothesis_probability"] = results.loc[
        cell_hashing_adata.obs_names, "negative_hypothesis_probability"
    ]
    cell_hashing_adata.obs["singlet_hypothesis_probability"] = results.loc[
        cell_hashing_adata.obs_names, "singlet_hypothesis_probability"
    ]
    cell_hashing_adata.obs["doublet_hypothesis_probability"] = results.loc[
        cell_hashing_adata.obs_names, "doublet_hypothesis_probability"
    ]

    cell_hashing_adata.obs["Classification"] = None
    cell_hashing_adata.obs.loc[
        cell_hashing_adata.obs["most_likely_hypothesis"] == 2, "Classification"
    ] = "Doublet"
    cell_hashing_adata.obs.loc[
        cell_hashing_adata.obs["most_likely_hypothesis"] == 0, "Classification"
    ] = "Negative"
    all_sings = cell_hashing_adata.obs["most_likely_hypothesis"] == 1
    singlet_sample_index = np.argmax(cell_hashing_adata.X[all_sings], axis=1)
    cell_hashing_adata.obs.loc[
        all_sings, "Classification"
    ] = cell_hashing_adata.var_names[singlet_sample_index]

    return cell_hashing_adata if not inplace else None


def plot_qc_checks_cell_hashing(
    cell_hashing_adata: anndata.AnnData, alpha: float = 0.05, fig_path: str = None
):
    """Plot HashSolo demultiplexing results

    Parameters
    ----------
    cell_hashing_adata : Anndata
        Anndata object filled only with hashing counts
    alpha : float
        Tranparency of scatterplot points
    fig_path : str
        Path to save figure
    Returns
    -------
    """
    import matplotlib

    matplotlib.use("Agg")
    import matplotlib.pyplot as plt

    cell_hashing_demultiplexing = cell_hashing_adata.obs
    cell_hashing_demultiplexing["log_counts"] = np.log(
        np.sum(cell_hashing_adata.X, axis=1)
    )
    number_of_clusters = (
        cell_hashing_demultiplexing["cluster_feature"].drop_duplicates().shape[0]
    )
    fig, all_axes = plt.subplots(
        number_of_clusters, 4, figsize=(40, 10 * number_of_clusters)
    )
    counter = 0
    for cluster_feature, group in cell_hashing_demultiplexing.groupby(
        "cluster_feature"
    ):
        if number_of_clusters > 1:
            axes = all_axes[counter]
        else:
            axes = all_axes

        ax = axes[0]
        ax.plot(
            group["log_counts"],
            group["negative_hypothesis_probability"],
            "bo",
            alpha=alpha,
        )
        ax.set_title("Probability of negative hypothesis vs log hashing counts")
        ax.set_ylabel("Probability of negative hypothesis")
        ax.set_xlabel("Log hashing counts")

        ax = axes[1]
        ax.plot(
            group["log_counts"],
            group["singlet_hypothesis_probability"],
            "bo",
            alpha=alpha,
        )
        ax.set_title("Probability of singlet hypothesis vs log hashing counts")
        ax.set_ylabel("Probability of singlet hypothesis")
        ax.set_xlabel("Log hashing counts")

        ax = axes[2]
        ax.plot(
            group["log_counts"],
            group["doublet_hypothesis_probability"],
            "bo",
            alpha=alpha,
        )
        ax.set_title("Probability of doublet hypothesis vs log hashing counts")
        ax.set_ylabel("Probability of doublet hypothesis")
        ax.set_xlabel("Log hashing counts")

        ax = axes[3]
        group["Classification"].value_counts().plot.bar(ax=ax)
        ax.set_title("Count of each samples classification")
        counter += 1
    plt.show()
    if fig_path is not None:
        fig.savefig(fig_path, dpi=300, format="pdf")


def main():
    usage = "hashsolo"
    parser = ArgumentParser(usage, formatter_class=ArgumentDefaultsHelpFormatter)

    parser.add_argument(
        dest="data_file", help="h5ad file containing cell hashing counts"
    )
    parser.add_argument(
        "-j",
        dest="model_json_file",
        default=None,
        help="json file to pass optional arguments",
    )
    parser.add_argument(
        "-o",
        dest="out_dir",
        default="hashsolo_output",
        help="Output directory for results",
    )
    parser.add_argument(
        "-c",
        dest="clustering_data",
        default=None,
        help="h5ad file with count transcriptional data to\
                    perform clustering on",
    )
    parser.add_argument(
        "-p",
        dest="pre_existing_clusters",
        default=None,
        help="column in cell_hashing_data_file.obs to \
                        specifying different cell types or clusters",
    )
    parser.add_argument(
        "-q",
        dest="plot_name",
        default="hashing_qc_plots.pdf",
        help="name of plot to output",
    )
    parser.add_argument(
        "-n",
        dest="number_of_noise_barcodes",
        default=None,
        help="Number of barcodes to use to create noise \
                        distribution",
    )

    args = parser.parse_args()

    model_json_file = args.model_json_file
    if model_json_file is not None:
        # read parameters
        with open(model_json_file) as model_json_open:
            params = json.load(model_json_open)
    else:
        params = {}
    data_file = args.data_file
    data_ext = os.path.splitext(data_file)[-1]
    if data_ext == ".h5ad":
        cell_hashing_adata = anndata.read(data_file)
    else:
        print("Unrecognized file format")

    if args.clustering_data is not None:
        clustering_data_file = args.clustering_data
        clustering_data_ext = os.path.splitext(clustering_data_file)[-1]
        if clustering_data_ext == ".h5ad":
            clustering_data = anndata.read(clustering_data_file)
        else:
            print("Unrecognized file format for clustering data")
    else:
        clustering_data = None

    if not os.path.isdir(args.out_dir):
        os.mkdir(args.out_dir)

    hashsolo(
        cell_hashing_adata,
        pre_existing_clusters=args.pre_existing_clusters,
        number_of_noise_barcodes=args.number_of_noise_barcodes,
        clustering_data=clustering_data,
        **params
    )
    cell_hashing_adata.write(os.path.join(args.out_dir, "hashsoloed.h5ad"))
    plot_qc_checks_cell_hashing(
        cell_hashing_adata, fig_path=os.path.join(args.out_dir, args.plot_name)
    )


###############################################################################
# __main__
###############################################################################


if __name__ == "__main__":
    main()



import numpy as np

from scipy.stats import multinomial
from sklearn.neighbors import NearestNeighbors


def knn_smooth_pred_class(
    X: np.ndarray,
    pred_class: np.ndarray,
    grouping: np.ndarray = None,
    k: int = 15,
) -> np.ndarray:
    """
    Smooths class predictions by taking the modal class from each cell's
    nearest neighbors.
    Parameters
    ----------
    X : np.ndarray
        [N, Features] embedding space for calculation of nearest neighbors.
    pred_class : np.ndarray
        [N,] array of unique class labels.
    groupings : np.ndarray
        [N,] unique grouping labels for i.e. clusters.
        if provided, only considers nearest neighbors *within the cluster*.
    k : int
        number of nearest neighbors to use for smoothing.
    Returns
    -------
    smooth_pred_class : np.ndarray
        [N,] unique class labels, smoothed by kNN.
    Examples
    --------
    >>> smooth_pred_class = knn_smooth_pred_class(
    ...     X = X,
    ...     pred_class = raw_predicted_classes,
    ...     grouping = louvain_cluster_groups,
    ...     k = 15,)
    Notes
    -----
    scNym classifiers do not incorporate neighborhood information.
    By using a simple kNN smoothing heuristic, we can leverage neighborhood
    information to improve classification performance, smoothing out cells
    that have an outlier prediction relative to their local neighborhood.
    """
    if grouping is None:
        # do not use a grouping to restrict local neighborhood
        # associations, create a universal pseudogroup `0`.
        grouping = np.zeros(X.shape[0])

    smooth_pred_class = np.zeros_like(pred_class)
    for group in np.unique(grouping):
        # identify only cells in the relevant group
        group_idx = np.where(grouping == group)[0].astype("int")
        X_group = X[grouping == group, :]
        # if there are < k cells in the group, change `k` to the
        # group size
        if X_group.shape[0] < k:
            k_use = X_group.shape[0]
        else:
            k_use = k
        # compute a nearest neighbor graph and identify kNN
        nns = NearestNeighbors(
            n_neighbors=k_use,
        ).fit(X_group)
        dist, idx = nns.kneighbors(X_group)

        # for each cell in the group, assign a class as
        # the majority class of the kNN
        for i in range(X_group.shape[0]):
            classes = pred_class[group_idx[idx[i, :]]]
            uniq_classes, counts = np.unique(classes, return_counts=True)
            maj_class = uniq_classes[int(np.argmax(counts))]
            smooth_pred_class[group_idx[i]] = maj_class
    return smooth_pred_class



#!/usr/bin/env python
import json
import os
import umap
import sys
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import pkg_resources

import numpy as np
from sklearn.metrics import *
from scipy.special import softmax
from scanpy import read_10x_mtx

import torch
from pytorch_lightning.callbacks.early_stopping import EarlyStopping

import scvi
from scvi.data import read_h5ad, read_loom, setup_anndata
from scvi.model import SCVI
from scvi.external import SOLO

from .utils import knn_smooth_pred_class

"""
solo.py

Simulate doublets, train a VAE, and then a classifier on top.
"""


###############################################################################
# main
###############################################################################


def main():
    usage = "solo"
    parser = ArgumentParser(usage, formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "-j",
        dest="model_json_file",
        help="json file to pass VAE parameters",
        required="--version" not in sys.argv,
    )
    parser.add_argument(
        "-d",
        dest="data_path",
        help="path to h5ad, loom, or 10x mtx dir cell by genes counts",
        required="--version" not in sys.argv,
    )
    parser.add_argument(
        "--set-reproducible-seed",
        dest="reproducible_seed",
        default=None,
        type=int,
        help="Reproducible seed, give an int to set seed",
    )
    parser.add_argument(
        "--doublet-depth",
        dest="doublet_depth",
        default=2.0,
        type=float,
        help="Depth multiplier for a doublet relative to the \
                        average of its constituents",
    )
    parser.add_argument(
        "-g", dest="gpu", default=True, action="store_true", help="Run on GPU"
    )
    parser.add_argument(
        "-a",
        dest="anndata_output",
        default=False,
        action="store_true",
        help="output modified anndata object with solo scores \
                        Only works for anndata",
    )
    parser.add_argument("-o", dest="out_dir", default="solo_out")
    parser.add_argument(
        "-r",
        dest="doublet_ratio",
        default=2,
        type=int,
        help="Ratio of doublets to true \
                        cells",
    )
    parser.add_argument(
        "-s",
        dest="seed",
        default=None,
        help="Path to previous solo output  \
                        directory. Seed VAE models with previously \
                        trained solo model. Directory structure is assumed to \
                        be the same as solo output directory structure. \
                        should at least have a vae.pt a pickled object of \
                        vae weights and a latent.npy an np.ndarray of the \
                        latents of your cells.",
    )
    parser.add_argument(
        "-e",
        dest="expected_number_of_doublets",
        help="Experimentally expected number of doublets",
        type=int,
        default=None,
    )
    parser.add_argument(
        "-p",
        dest="plot",
        default=False,
        action="store_true",
        help="Plot outputs for solo",
    )
    parser.add_argument(
        "-recalibrate_scores",
        dest="recalibrate_scores",
        default=False,
        action="store_true",
        help="Recalibrate doublet scores (not recommended anymore)",
    )
    parser.add_argument(
        "--version",
        dest="version",
        default=False,
        action="store_true",
        help="Get version of solo-sc",
    )
    args = parser.parse_args()

    if args.version:
        version = pkg_resources.require("solo-sc")[0].version
        print(f"Current version of solo-sc is {version}")
        if args.model_json_file is None or args.data_path is None:
            print("Json or data path not give exiting solo")
            sys.exit()

    model_json_file = args.model_json_file
    data_path = args.data_path
    if args.gpu and not torch.cuda.is_available():
        args.gpu = torch.cuda.is_available()
        print("Cuda is not available, switching to cpu running!")

    if not os.path.isdir(args.out_dir):
        os.mkdir(args.out_dir)

    if args.reproducible_seed is not None:
        scvi.settings.seed = args.reproducible_seed
    else:
        scvi.settings.seed = np.random.randint(10000)

    ##################################################
    # data

    # read loom/anndata
    data_ext = os.path.splitext(data_path)[-1]
    if data_ext == ".loom":
        scvi_data = read_loom(data_path)
    elif data_ext == ".h5ad":
        scvi_data = read_h5ad(data_path)
    elif os.path.isdir(data_path):
        scvi_data = read_10x_mtx(path=data_path)
        cell_umi_depth = scvi_data.X.sum(axis=1)
        fifth, ninetyfifth = np.percentile(cell_umi_depth, [5, 95])
        min_cell_umi_depth = np.min(cell_umi_depth)
        max_cell_umi_depth = np.max(cell_umi_depth)
        if fifth * 10 < ninetyfifth:
            print(
                """WARNING YOUR DATA HAS A WIDE RANGE OF CELL DEPTHS.
            PLEASE MANUALLY REVIEW YOUR DATA"""
            )
        print(
            f"Min cell depth: {min_cell_umi_depth}, Max cell depth: {max_cell_umi_depth}"
        )
    else:
        msg = f"{data_path} is not a recognized format.\n"
        msg += "must be one of {h5ad, loom, 10x mtx dir}"
        raise TypeError(msg)

    num_cells, num_genes = scvi_data.X.shape

    # check for parameters
    if not os.path.exists(model_json_file):
        raise FileNotFoundError(f"{model_json_file} does not exist.")
    # read parameters
    with open(model_json_file, "r") as model_json_open:
        params = json.load(model_json_open)

    # set VAE params
    vae_params = {}
    for par in ["n_hidden", "n_latent", "n_layers", "dropout_rate", "ignore_batch"]:
        if par in params:
            vae_params[par] = params[par]

    # training parameters
    batch_key = params.get("batch_key", None)
    batch_size = params.get("batch_size", 128)
    valid_pct = params.get("valid_pct", 0.1)
    learning_rate = params.get("learning_rate", 1e-3)
    stopping_params = {"patience": params.get("patience", 8), "min_delta": 0}

    # protect against single example batch
    while num_cells % batch_size == 1:
        batch_size = int(np.round(1.25 * batch_size))
        print("Increasing batch_size to %d to avoid single example batch." % batch_size)

    scvi.settings.batch_size = batch_size
    ##################################################
    # SCVI
    setup_anndata(scvi_data, batch_key=batch_key)
    vae = SCVI(
        scvi_data,
        gene_likelihood="nb",
        log_variational=True,
        **vae_params,
        use_observed_lib_size=False,
    )

    if args.seed:
        vae = vae.load(os.path.join(args.seed, "vae"), use_gpu=args.gpu)
    else:
        scvi_callbacks = []
        scvi_callbacks += [
            EarlyStopping(
                monitor="reconstruction_loss_validation", mode="min", **stopping_params
            )
        ]
        plan_kwargs = {
            "reduce_lr_on_plateau": True,
            "lr_factor": 0.1,
            "lr": 1e-2,
            "lr_patience": 10,
            "lr_threshold": 0,
            "lr_min": 1e-4,
            "lr_scheduler_metric": "reconstruction_loss_validation",
        }

        vae.train(
            max_epochs=2000,
            validation_size=valid_pct,
            check_val_every_n_epoch=5,
            plan_kwargs=plan_kwargs,
            callbacks=scvi_callbacks,
        )
        # save VAE
        vae.save(os.path.join(args.out_dir, "vae"))

    latent = vae.get_latent_representation()
    # save latent representation
    np.save(os.path.join(args.out_dir, "latent.npy"), latent.astype("float32"))

    ##################################################
    # classifier

    # model
    # todo add doublet ratio
    solo = SOLO.from_scvi_model(vae, doublet_ratio=args.doublet_ratio)
    solo.train(
        2000,
        lr=learning_rate,
        train_size=0.9,
        check_val_every_n_epoch=5,
        early_stopping_patience=6,
    )
    solo.train(
        2000,
        lr=learning_rate * 0.1,
        train_size=0.9,
        check_val_every_n_epoch=1,
        early_stopping_patience=30,
        callbacks=[],
    )
    solo.save(os.path.join(args.out_dir, "classifier"))

    logit_predictions = solo.predict(include_simulated_doublets=True)

    is_doublet_known = solo.adata.obs._solo_doub_sim == "doublet"
    is_doublet_pred = logit_predictions.idxmin(axis=1) == "singlet"

    validation_is_doublet_known = is_doublet_known[solo.validation_indices]
    validation_is_doublet_pred = is_doublet_pred[solo.validation_indices]
    training_is_doublet_known = is_doublet_known[solo.train_indices]
    training_is_doublet_pred = is_doublet_pred[solo.train_indices]

    valid_as = accuracy_score(validation_is_doublet_known, validation_is_doublet_pred)
    valid_roc = roc_auc_score(validation_is_doublet_known, validation_is_doublet_pred)
    valid_ap = average_precision_score(
        validation_is_doublet_known, validation_is_doublet_pred
    )

    train_as = accuracy_score(training_is_doublet_known, training_is_doublet_pred)
    train_roc = roc_auc_score(training_is_doublet_known, training_is_doublet_pred)
    train_ap = average_precision_score(
        training_is_doublet_known, training_is_doublet_pred
    )

    print(f"Training results")
    print(f"AUROC: {train_roc}, Accuracy: {train_as}, Average precision: {train_ap}")

    print(f"Validation results")
    print(f"AUROC: {valid_roc}, Accuracy: {valid_as}, Average precision: {valid_ap}")

    # write predictions
    # softmax predictions
    softmax_predictions = softmax(logit_predictions, axis=1)
    doublet_score = softmax_predictions.loc[:, "doublet"]

    np.save(
        os.path.join(args.out_dir, "no_updates_softmax_scores.npy"),
        doublet_score[:num_cells],
    )
    np.savetxt(
        os.path.join(args.out_dir, "no_updates_softmax_scores.csv"),
        doublet_score[:num_cells],
        delimiter=",",
    )
    np.save(
        os.path.join(args.out_dir, "no_updates_softmax_scores_sim.npy"),
        doublet_score[num_cells:],
    )

    # logit predictions
    logit_doublet_score = logit_predictions.loc[:, "doublet"]
    np.save(
        os.path.join(args.out_dir, "logit_scores.npy"), logit_doublet_score[:num_cells]
    )
    np.savetxt(
        os.path.join(args.out_dir, "logit_scores.csv"),
        logit_doublet_score[:num_cells],
        delimiter=",",
    )
    np.save(
        os.path.join(args.out_dir, "logit_scores_sim.npy"),
        logit_doublet_score[num_cells:],
    )

    # update threshold as a function of Solo's estimate of the number of
    # doublets
    # essentially a log odds update
    # TODO put in a function
    # currently overshrinking softmaxes
    diff = np.inf
    counter_update = 0
    solo_scores = doublet_score[:num_cells]
    logit_scores = logit_doublet_score[:num_cells]
    d_s = args.doublet_ratio / (args.doublet_ratio + 1)
    if args.recalibrate_scores:
        while (diff > 0.01) | (counter_update < 5):

            # calculate log odds calibration for logits
            d_o = np.mean(solo_scores)
            c = np.log(d_o / (1 - d_o)) - np.log(d_s / (1 - d_s))

            # update solo scores
            solo_scores = 1 / (1 + np.exp(-(logit_scores + c)))

            # update while conditions
            diff = np.abs(d_o - np.mean(solo_scores))
            counter_update += 1

    np.save(os.path.join(args.out_dir, "softmax_scores.npy"), solo_scores)
    np.savetxt(
        os.path.join(args.out_dir, "softmax_scores.csv"), solo_scores, delimiter=","
    )

    if args.expected_number_of_doublets is not None:
        k = len(solo_scores) - args.expected_number_of_doublets
        if args.expected_number_of_doublets / len(solo_scores) > 0.5:
            print(
                """Make sure you actually expect more than half your cells
                   to be doublets. If not change your
                   -e parameter value"""
            )
        assert k > 0
        idx = np.argpartition(solo_scores, k)
        threshold = np.max(solo_scores[idx[:k]])
        is_solo_doublet = solo_scores > threshold
    else:
        is_solo_doublet = solo_scores > 0.5

    np.save(os.path.join(args.out_dir, "is_doublet.npy"), is_solo_doublet[:num_cells])
    np.savetxt(
        os.path.join(args.out_dir, "is_doublet.csv"),
        is_solo_doublet[:num_cells],
        delimiter=",",
    )

    np.save(
        os.path.join(args.out_dir, "is_doublet_sim.npy"), is_solo_doublet[num_cells:]
    )

    np.save(os.path.join(args.out_dir, "preds.npy"), is_doublet_pred[:num_cells])
    np.savetxt(
        os.path.join(args.out_dir, "preds.csv"),
        is_doublet_pred[:num_cells],
        delimiter=",",
    )

    smoothed_preds = knn_smooth_pred_class(
        X=latent, pred_class=is_doublet_pred[:num_cells]
    )
    np.save(os.path.join(args.out_dir, "smoothed_preds.npy"), smoothed_preds)

    if args.anndata_output and data_ext == ".h5ad":
        scvi_data.obs["is_doublet"] = is_solo_doublet[:num_cells].values.astype(bool)
        scvi_data.obs["logit_scores"] = logit_doublet_score[:num_cells].values.astype(
            float
        )
        scvi_data.obs["softmax_scores"] = solo_scores[:num_cells].values.astype(float)
        scvi_data.write(os.path.join(args.out_dir, "soloed.h5ad"))

    if args.plot:
        import matplotlib

        matplotlib.use("Agg")
        import matplotlib.pyplot as plt
        import seaborn as sns

        train_solo_scores = doublet_score[solo.train_indices]
        validation_solo_scores = doublet_score[solo.validation_indices]

        train_fpr, train_tpr, _ = roc_curve(
            training_is_doublet_known, train_solo_scores
        )
        val_fpr, val_tpr, _ = roc_curve(
            validation_is_doublet_known, validation_solo_scores
        )

        # plot ROC
        plt.figure()
        plt.plot(train_fpr, train_tpr, label="Train")
        plt.plot(val_fpr, val_tpr, label="Validation")
        plt.gca().set_xlabel("False positive rate")
        plt.gca().set_ylabel("True positive rate")
        plt.legend()
        plt.savefig(os.path.join(args.out_dir, "roc.pdf"))
        plt.close()

        train_precision, train_recall, _ = precision_recall_curve(
            training_is_doublet_known, train_solo_scores
        )
        val_precision, val_recall, _ = precision_recall_curve(
            validation_is_doublet_known, validation_solo_scores
        )
        # plot accuracy
        plt.figure()
        plt.plot(train_recall, train_precision, label="Train")
        plt.plot(val_recall, val_precision, label="Validation")
        plt.gca().set_xlabel("Recall")
        plt.gca().set_ylabel("pytPrecision")
        plt.legend()
        plt.savefig(os.path.join(args.out_dir, "precision_recall.pdf"))
        plt.close()

        # plot distributions
        obs_indices = solo.validation_indices[solo.validation_indices < num_cells]
        sim_indices = solo.validation_indices[solo.validation_indices > num_cells]

        plt.figure()
        sns.displot(doublet_score[sim_indices], label="Simulated")
        sns.displot(doublet_score[obs_indices], label="Observed")
        plt.legend()
        plt.savefig(os.path.join(args.out_dir, "sim_vs_obs_dist.pdf"))
        plt.close()

        plt.figure()
        sns.distplot(solo_scores[:num_cells], label="Observed (transformed)")
        plt.legend()
        plt.savefig(os.path.join(args.out_dir, "real_cells_dist.pdf"))
        plt.close()

        scvi_umap = umap.UMAP(n_neighbors=16).fit_transform(latent)
        fig, ax = plt.subplots(1, 1, figsize=(10, 10))
        ax.scatter(
            scvi_umap[:, 0],
            scvi_umap[:, 1],
            c=doublet_score[:num_cells],
            s=8,
            cmap="GnBu",
        )

        ax.set_xlabel("UMAP 1")
        ax.set_ylabel("UMAP 2")
        fig.savefig(os.path.join(args.out_dir, "umap_solo_scores.pdf"))


###############################################################################
# __main__
###############################################################################


if __name__ == "__main__":
    main()



import sys
from pathlib import Path
from setuptools import setup, find_packages

if sys.version_info < (3,):
    sys.exit("solo requires Python >= 3.6")

try:
    from solo import __author__, __email__
except ImportError:  # Deps not yet installed
    __author__ = __email__ = ""


setup(
    name="solo-sc",
    version="1.0",
    description="Neural network classifiers for doublets",
    long_description=Path("README.md").read_text("utf-8"),
    long_description_content_type="text/markdown",
    url="http://github.com/calico/solo",
    download_url="https://github.com/calico/solo/archive/1.0.tar.gz",
    author=__author__,
    author_email=__email__,
    license="Apache",
    python_requires=">=3.6",
    install_requires=[
        l.strip() for l in Path("requirements.txt").read_text("utf-8").splitlines()
    ],
    packages=find_packages(exclude="testdata"),
    entry_points=dict(
        console_scripts=["solo=solo.solo:main", "hashsolo=solo.hashsolo:main"],
    ),
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
)
